@@ -14,30 +14,38 @@
 
 #include "dnsmasq.h"
 
-static struct irec *add_iface(struct irec *list, char *name, union mysockaddr *addr, 
-			      struct iname *names, struct iname *addrs, 
-			      struct iname *except)
+static struct irec *add_iface(struct daemon *daemon, struct irec *list, char *name, union mysockaddr *addr) 
 {
   struct irec *iface;
   struct iname *tmp;
   
   /* check blacklist */
-  if (except)
-    for (tmp = except; tmp; tmp = tmp->next)
+  if (daemon->if_except)
+    for (tmp = daemon->if_except; tmp; tmp = tmp->next)
       if (tmp->name && strcmp(tmp->name, name) == 0)
-	return list;
-  
+	{
+	  /* record address of named interfaces, for TCP access control */
+	  tmp->addr = *addr;
+	  return list;
+	}
+
   /* we may need to check the whitelist */
-  if (names || addrs)
+  if (daemon->if_names || daemon->if_addrs)
     { 
-      for (tmp = names; tmp; tmp = tmp->next)
+      int found = 0;
+
+      for (tmp = daemon->if_names; tmp; tmp = tmp->next)
 	if (tmp->name && (strcmp(tmp->name, name) == 0))
-	  break;
-      if (!tmp) 
-	for (tmp = addrs; tmp; tmp = tmp->next)
-	  if (sockaddr_isequal(&tmp->addr, addr))
-	    break;
-      if (!tmp) 
+	  {
+	    tmp->addr = *addr;
+	    found = tmp->used = 1;
+	  }
+
+      for (tmp = daemon->if_addrs; tmp; tmp = tmp->next)
+	if (sockaddr_isequal(&tmp->addr, addr))
+	  found = tmp->used = 1;
+      
+      if (!found) 
 	return list;
     }
   
@@ -57,10 +65,7 @@ static struct irec *add_iface(struct irec *list, char *name, union mysockaddr *a
 }
 
 
-struct irec *enumerate_interfaces(struct iname **names,
-				  struct iname **addrs,
-				  struct iname *except,
-				  int port)
+struct irec *enumerate_interfaces(struct daemon *daemon)
 {
   struct irec *iface = NULL;
   char *buf, *ptr;
@@ -72,7 +77,7 @@ struct irec *enumerate_interfaces(struct iname **names,
   
   if (fd == -1)
     die ("cannot create socket to enumerate interfaces: %s", NULL);
-      
+        
   while (1)
      {
        buf = safe_malloc(len);
@@ -116,7 +121,7 @@ struct irec *enumerate_interfaces(struct iname **names,
       if (ifr->ifr_addr.sa_family == AF_INET)
 	{
 	  addr.in = *((struct sockaddr_in *) &ifr->ifr_addr);
-	  addr.in.sin_port = htons(port);
+	  addr.in.sin_port = htons(daemon->port);
 	}
 #ifdef HAVE_IPV6
       else if (ifr->ifr_addr.sa_family == AF_INET6)
@@ -126,7 +131,7 @@ struct irec *enumerate_interfaces(struct iname **names,
 #else
 	  addr.in6 = *((struct sockaddr_in6 *) &ifr->ifr_addr);
 #endif
-	  addr.in6.sin6_port = htons(port);
+	  addr.in6.sin6_port = htons(daemon->port);
 	  addr.in6.sin6_flowinfo = htonl(0);
 	}
 #endif
@@ -137,18 +142,27 @@ struct irec *enumerate_interfaces(struct iname **names,
 	die("ioctl error getting interface flags: %m", NULL);
 
       /* If we are restricting the set of interfaces to use, make
-	 sure that loopback interfaces are in that set. Note that
-	 this is done as addresses rather than interface names so
-	 as not to confuse the no-IPRECVIF workaround on the DHCP code */
-      if (*names && (ifr->ifr_flags & IFF_LOOPBACK))
+	 sure that loopback interfaces are in that set. */
+      if (daemon->if_names && (ifr->ifr_flags & IFF_LOOPBACK))
 	{
-	  struct iname *lo = safe_malloc(sizeof(struct iname));
-	  lo->addr = addr;
-	  lo->next = *addrs;
-	  *addrs = lo;
+	  struct iname *lo;
+	  for (lo = daemon->if_names; lo; lo = lo->next)
+	    if (lo->name && strcmp(lo->name, ifr->ifr_name) == 0)
+	      {
+		lo->isloop = 1;
+		break;
+	      }
+	  if (!lo)
+	    {
+	      lo = safe_malloc(sizeof(struct iname));
+	      lo->name = safe_string_alloc(ifr->ifr_name);
+	      lo->isloop = lo->used = 1;
+	      lo->next = daemon->if_names;
+	      daemon->if_names = lo;
+	    }
 	}
 
-      iface = add_iface(iface, ifr->ifr_name, &addr, *names, *addrs, except);
+      iface = add_iface(daemon, iface, ifr->ifr_name, &addr);
 	
 #if defined(HAVE_LINUX_IPV6_PROC) && defined(HAVE_IPV6)
       /* IPv6 addresses don't seem to work with SIOCGIFCONF. Barf */
@@ -179,7 +193,7 @@ struct irec *enumerate_interfaces(struct iname **names,
 			sscanf(addrstring+i+i, "%02x", &byte);
 			addr6p[i] = byte;
 		      }
-		    addr6.in6.sin6_port = htons(port);
+		    addr6.in6.sin6_port = htons(daemon->port);
 		    addr6.in6.sin6_flowinfo = htonl(0);
 		    addr6.in6.sin6_scope_id = htonl(scope);
 		    
@@ -192,17 +206,7 @@ struct irec *enumerate_interfaces(struct iname **names,
 	  }
 	
 	if (found)
-	  {
-	    if (*names && (ifr->ifr_flags & IFF_LOOPBACK))
-	      {
-		struct iname *lo = safe_malloc(sizeof(struct iname));
-		lo->addr = addr6;
-		lo->next = *addrs;
-		*addrs = lo;
-	      }
-	    
-	    iface = add_iface(iface, ifr->ifr_name, &addr6, *names, *addrs, except);
-	  }
+	  iface = add_iface(daemon, iface, ifr->ifr_name, &addr6);
       }
 #endif /* LINUX */
     }
@@ -218,36 +222,14 @@ struct irec *enumerate_interfaces(struct iname **names,
   return iface;
 }
 
-struct listener *create_wildcard_listeners(int port)
+#ifdef HAVE_IPV6
+static int create_ipv6_listener(struct listener **link, int port)
 {
   union mysockaddr addr;
+  int tcpfd, fd, flags, save;
+  struct listener *l;
   int opt = 1;
-  struct listener *listen;
-#ifdef HAVE_IPV6
-  int fd;
-#endif
 
-  addr.in.sin_family = AF_INET;
-  addr.in.sin_addr.s_addr = INADDR_ANY;
-  addr.in.sin_port = htons(port);
-#ifdef HAVE_SOCKADDR_SA_LEN
-  addr.in.sin_len = sizeof(struct sockaddr_in);
-#endif
-  listen = safe_malloc(sizeof(struct listener));
-  if ((listen->fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
-    die("failed to create socket: %s", NULL);
-  if (setsockopt(listen->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-#if defined(IP_PKTINFO) 
-      setsockopt(listen->fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
-#elif defined(IP_RECVDSTADDR) && defined(IP_RECVIF)
-      setsockopt(listen->fd, IPPROTO_IP, IP_RECVDSTADDR, &opt, sizeof(opt)) == -1 ||
-      setsockopt(listen->fd, IPPROTO_IP, IP_RECVIF, &opt, sizeof(opt)) == -1 ||
-#endif 
-      bind(listen->fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1)
-    die("failed to bind socket: %s", NULL);
-  listen->next = NULL;
-  listen->family = AF_INET;
-#ifdef HAVE_IPV6
   addr.in6.sin6_family = AF_INET6;
   addr.in6.sin6_addr = in6addr_any;
   addr.in6.sin6_port = htons(port);
@@ -255,53 +237,152 @@ struct listener *create_wildcard_listeners(int port)
 #ifdef HAVE_SOCKADDR_SA_LEN
   addr.in6.sin6_len = sizeof(struct sockaddr_in6);
 #endif
+
+  /* No error of the kernel doesn't support IPv6 */
   if ((fd = socket(AF_INET6, SOCK_DGRAM, 0)) == -1)
-    {
-      if (errno != EPROTONOSUPPORT &&
-	  errno != EAFNOSUPPORT &&
-	  errno != EINVAL)
-	die("failed to create IPv6 socket: %s", NULL);
-    }
-  else
-    {
-      listen->next = safe_malloc(sizeof(struct listener));
-      listen->next->fd = fd;
-      listen->next->family = AF_INET6;
-      listen->next->next = NULL;
-      if (setsockopt(listen->next->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-	  setsockopt(listen->next->fd, IPV6_LEVEL, IPV6_PKTINFO, &opt, sizeof(opt)) == -1 ||
-	  bind(listen->next->fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1)
-	die("failed to bind IPv6 socket: %s", NULL);
-    }
-#endif
+    return (errno == EPROTONOSUPPORT ||
+	    errno == EAFNOSUPPORT ||
+	    errno == EINVAL);
   
-  return listen;
+  if ((tcpfd = socket(AF_INET6, SOCK_STREAM, 0)) == -1)
+    {
+      save = errno;
+      close(fd);
+      errno = save;
+      return 0;
+    }
+  
+  if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+      setsockopt(tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+      setsockopt(fd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1 ||
+      setsockopt(tcpfd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1 ||
+      (flags = fcntl(tcpfd, F_GETFL, 0)) == -1 ||
+      fcntl(tcpfd, F_SETFL, flags | O_NONBLOCK) == -1 ||
+#ifdef IPV6_RECVPKTINFO
+      setsockopt(fd, IPV6_LEVEL, IPV6_RECVPKTINFO, &opt, sizeof(opt)) == -1 ||
+#else
+      setsockopt(fd, IPV6_LEVEL, IPV6_PKTINFO, &opt, sizeof(opt)) == -1 ||
+#endif
+      bind(tcpfd, (struct sockaddr *)&addr, sa_len(&addr)) == -1 ||
+      listen(tcpfd, 5) == -1 ||
+      bind(fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1) 
+    {
+      save = errno;
+      close(fd);
+      close(tcpfd);
+      errno = save;
+      return 0;
+    }
+  
+  l = safe_malloc(sizeof(struct listener));
+  l->fd = fd;
+  l->tcpfd = tcpfd;
+  l->family = AF_INET6;
+  l->next = NULL;
+  *link = l;
+  
+  return 1;
+}
+#endif
+
+struct listener *create_wildcard_listeners(int port)
+{
+#if !(defined(IP_PKTINFO) || (defined(IP_RECVDSTADDR) && defined(IP_RECVIF) && defined(IP_SENDSRCADDR)))
+  return NULL;
+#else
+  union mysockaddr addr;
+  int opt = 1;
+  struct listener *l, *l6 = NULL;
+  int flags;
+  int tcpfd, fd;
+
+  addr.in.sin_family = AF_INET;
+  addr.in.sin_addr.s_addr = INADDR_ANY;
+  addr.in.sin_port = htons(port);
+#ifdef HAVE_SOCKADDR_SA_LEN
+  addr.in.sin_len = sizeof(struct sockaddr_in);
+#endif
+
+  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
+    return NULL;
+  
+  if ((tcpfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
+    {
+      close (fd);
+      return NULL;
+    }
+  
+  if (setsockopt(tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+      bind(tcpfd, (struct sockaddr *)&addr, sa_len(&addr)) == -1 ||
+      listen(tcpfd, 5) == -1 ||
+      (flags = fcntl(tcpfd, F_GETFL, 0)) == -1 ||
+      fcntl(tcpfd, F_SETFL, flags | O_NONBLOCK) == -1 ||
+#ifdef HAVE_IPV6
+      !create_ipv6_listener(&l6, port) ||
+#endif
+      setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+#if defined(IP_PKTINFO) 
+      setsockopt(fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
+#elif defined(IP_RECVDSTADDR) && defined(IP_RECVIF)
+      setsockopt(fd, IPPROTO_IP, IP_RECVDSTADDR, &opt, sizeof(opt)) == -1 ||
+      setsockopt(fd, IPPROTO_IP, IP_RECVIF, &opt, sizeof(opt)) == -1 ||
+#endif 
+      bind(fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1)
+    {
+      close(fd);
+      close(tcpfd);
+      return NULL;
+    }
+  
+  l = safe_malloc(sizeof(struct listener));
+  l->family = AF_INET;
+  l->fd = fd;
+  l->tcpfd = tcpfd;
+  l->next = l6;
+
+  return l;
+
+#endif
 }
 
-struct listener *create_bound_listeners(struct irec *interfaces)
+struct listener *create_bound_listeners(struct irec *interfaces, int port)
 {
 
   struct listener *listeners = NULL;
   struct irec *iface;
-  int opt = 1;
+  int flags = port, opt = 1;
+  
+  /* Create bound listeners only for IPv4, IPv6 always binds the wildcard */
+
+#ifdef HAVE_IPV6
+  if (!create_ipv6_listener(&listeners, port))
+    die("failed to to create listening socket: %s", NULL);
+#endif
 
   for (iface = interfaces ;iface; iface = iface->next)
-    {
-      struct listener *new = safe_malloc(sizeof(struct listener));
-      new->family = iface->addr.sa.sa_family;
-      new->next = listeners;
-      listeners = new;
-      if ((new->fd = socket(iface->addr.sa.sa_family, SOCK_DGRAM, 0)) == -1)
-	die("failed to create socket: %s", NULL);
-      if (setsockopt(new->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-	  bind(new->fd, &iface->addr.sa, sa_len(&iface->addr)) == -1)
-	die("failed to bind socket: %s", NULL);
-    }
-
+    if (iface->addr.sa.sa_family == AF_INET)
+      {
+	struct listener *new = safe_malloc(sizeof(struct listener));
+	new->family = iface->addr.sa.sa_family;
+	new->next = listeners;
+	listeners = new;
+	if ((new->tcpfd = socket(iface->addr.sa.sa_family, SOCK_STREAM, 0)) == -1 ||
+	    (new->fd = socket(iface->addr.sa.sa_family, SOCK_DGRAM, 0)) == -1 ||
+	    setsockopt(new->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	    setsockopt(new->tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	    /* See Stevens 16.6 */
+	    (flags = fcntl(new->tcpfd, F_GETFL, 0)) == -1 ||
+	    fcntl(new->tcpfd, F_SETFL, flags | O_NONBLOCK) == -1 ||
+	    bind(new->tcpfd, &iface->addr.sa, sa_len(&iface->addr)) == -1 ||
+	    bind(new->fd, &iface->addr.sa, sa_len(&iface->addr)) == -1 ||
+	    listen(new->tcpfd, 5) == -1)
+	  die("failed to to create listening socket: %s", NULL);
+      }
+  
   return listeners;
 }
 
-static struct serverfd *allocate_sfd(union mysockaddr *addr, struct serverfd **sfds)
+struct serverfd *allocate_sfd(union mysockaddr *addr, struct serverfd **sfds)
 {
   struct serverfd *sfd;
   
@@ -337,17 +418,19 @@ static struct serverfd *allocate_sfd(union mysockaddr *addr, struct serverfd **s
   return sfd;
 }
 
-struct server *check_servers(struct server *new, struct irec *interfaces, struct serverfd **sfds)
+void check_servers(struct daemon *daemon, struct irec *interfaces)
 {
   char addrbuff[ADDRSTRLEN];
   struct irec *iface;
-  struct server *tmp, *ret = NULL;
+  struct server *new, *tmp, *ret = NULL;
   int port = 0;
 
   /* forward table rules reference servers, so have to blow them away */
   forward_init(0);
   
-  for (;new; new = tmp)
+  daemon->last_server = NULL;
+  
+  for (new = daemon->servers; new; new = tmp)
     {
       tmp = new->next;
       
@@ -379,7 +462,7 @@ struct server *check_servers(struct server *new, struct irec *interfaces, struct
 	    }
 	  
 	  /* Do we need a socket set? */
-	  if (!new->sfd && !(new->sfd = allocate_sfd(&new->source_addr, sfds)))
+	  if (!new->sfd && !(new->sfd = allocate_sfd(&new->source_addr, &daemon->sfds)))
 	    {
 	      syslog(LOG_WARNING, 
 		     "ignoring nameserver %s - cannot make/bind socket: %m", addrbuff);
@@ -409,15 +492,16 @@ struct server *check_servers(struct server *new, struct irec *interfaces, struct
 	syslog(LOG_INFO, "using nameserver %s#%d", addrbuff, port); 
     }
   
- return ret;
+  daemon->servers = ret;
 }
   
-struct server *reload_servers(char *fname, char *buff, struct server *serv, int query_port)
+void reload_servers(char *fname, struct daemon *daemon)
 {
   FILE *f;
   char *line;
   struct server *old_servers = NULL;
   struct server *new_servers = NULL;
+  struct server *serv = daemon->servers;
 
   /* move old servers to free list - we can reuse the memory 
      and not risk malloc if there are the same or fewer new servers. 
@@ -447,7 +531,7 @@ struct server *reload_servers(char *fname, char *buff, struct server *serv, int
   else
     {
       syslog(LOG_INFO, "reading %s", fname);
-      while ((line = fgets(buff, MAXDNAME, f)))
+      while ((line = fgets(daemon->namebuff, MAXDNAME, f)))
 	{
 	  union  mysockaddr addr, source_addr;
 	  char *token = strtok(line, " \t\n\r");
@@ -455,7 +539,7 @@ struct server *reload_servers(char *fname, char *buff, struct server *serv, int
 	  
 	  if (!token || strcmp(token, "nameserver") != 0)
 	    continue;
-	  if (!(token = strtok(NULL, " \t\n")))
+	  if (!(token = strtok(NULL, " \t\n\r")))
 	    continue;
 	  
 #ifdef HAVE_IPV6
@@ -470,7 +554,7 @@ struct server *reload_servers(char *fname, char *buff, struct server *serv, int
 	      source_addr.in.sin_family = addr.in.sin_family = AF_INET;
 	      addr.in.sin_port = htons(NAMESERVER_PORT);
 	      source_addr.in.sin_addr.s_addr = INADDR_ANY;
-	      source_addr.in.sin_port = htons(query_port);
+	      source_addr.in.sin_port = htons(daemon->query_port);
 	    }
 #ifdef HAVE_IPV6
 	  else if (inet_pton(AF_INET6, token, &addr.in6.sin6_addr))
@@ -482,7 +566,7 @@ struct server *reload_servers(char *fname, char *buff, struct server *serv, int
 	      addr.in6.sin6_port = htons(NAMESERVER_PORT);
 	      source_addr.in6.sin6_flowinfo = addr.in6.sin6_flowinfo = htonl(0);
 	      source_addr.in6.sin6_addr = in6addr_any;
-	      source_addr.in6.sin6_port = htons(query_port);
+	      source_addr.in6.sin6_port = htons(daemon->query_port);
 	    }
 #endif /* IPV6 */
 	  else
@@ -518,7 +602,7 @@ struct server *reload_servers(char *fname, char *buff, struct server *serv, int
       old_servers = tmp;
     }
 
-  return new_servers;
+  daemon->servers = new_servers;
 }
 
 
