@@ -21,7 +21,7 @@ struct myoption {
   int val;
 };
 
-#define OPTSTRING "ZDNLERzowefnbvhdqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:"
+#define OPTSTRING "ZDNLERzowefnbvhdkqr:m:p:c:l:s:i:t:u:g:a:x:S:C:A:T:H:Q:I:B:F:G:O:M:X:V:U:j:P:"
 
 static struct myoption opts[] = { 
   {"version", 0, 0, 'v'},
@@ -72,6 +72,8 @@ static struct myoption opts[] = {
   {"alias", 1, 0, 'V' },
   {"dhcp-vendorclass", 1, 0, 'U'},
   {"dhcp-userclass", 1, 0, 'j'},
+  {"edns-packet-max", 1, 0, 'P'},
+  {"keep-in-foreground", 0, 0, 'k'},
   {0, 0, 0, 0}
 };
 
@@ -88,6 +90,7 @@ static struct optflags optmap[] = {
   { 'h', OPT_NO_HOSTS },
   { 'n', OPT_NO_POLL },
   { 'd', OPT_DEBUG },
+  { 'k', OPT_NO_FORK },
   { 'o', OPT_ORDER },
   { 'R', OPT_NO_RESOLV },
   { 'E', OPT_EXPAND },
@@ -123,6 +126,7 @@ static char *usage =
 "-i, --interface=interface           Specify interface(s) to listen on.\n"
 "-I, --except-interface=int          Specify interface(s) NOT to listen on.\n"
 "-j, --dhcp-userclass=<id>,<class>   Map DHCP user class to option set.\n"
+"-k, --keep-in-foreground            Do NOT fork into the background, do NOT run in debug mode.\n"
 "-l, --dhcp-leasefile=path           Specify where to store DHCP leases (defaults to " LEASEFILE ").\n"
 "-L, --localmx                       Return MX records for local hosts.\n"
 "-m, --mx-host=host_name             Specify the MX name to reply to.\n"
@@ -132,6 +136,7 @@ static char *usage =
 "-o, --strict-order                  Use nameservers strictly in the order given in " RESOLVFILE ".\n"
 "-O, --dhcp-option=<optspec>         Set extra options to be set to DHCP clients.\n"
 "-p, --port=number                   Specify port to listen for DNS requests on (defaults to 53).\n"
+"-P, --edns-packet-max=<size>        Maximum supported UDP packet size for EDNS.0 (defaults to %d).\n"
 "-q, --log-queries                   Log queries.\n"
 "-Q, --query-port=number             Force the originating port for upstream queries.\n"
 "-R, --no-resolv                     Do NOT read resolv.conf.\n"
@@ -143,7 +148,7 @@ static char *usage =
 "-T, --local-ttl=time                Specify time-to-live in seconds for replies from /etc/hosts.\n"
 "-u, --user=username                 Change to this user after startup. (defaults to " CHUSER ").\n" 
 "-U, --dhcp-vendorclass=<id>,<class> Map DHCP vendor class to option set.\n"
-"-v, --version                       Display dnsmasq version.\n"
+"-v, --version                       Display dnsmasq version and copyright information.\n"
 "-V, --alias=addr,addr,mask          Translate IPv4 addresses from upstream servers.\n"
 "-w, --help                          Display this message.\n"
 "-x, --pid-file=path                 Specify path of PID file. (defaults to " RUNFILE ").\n"
@@ -153,28 +158,37 @@ static char *usage =
 "\n";
 
 
-unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **resolv_files, 
-			char **mxname, char **mxtarget, char **lease_file, 
-			char **username, char **groupname, char **domain_suffix, char **runfile, 
-			struct iname **if_names, struct iname **if_addrs, struct iname **if_except,
-			struct bogus_addr **bogus_addr, struct server **serv_addrs, int *cachesize, int *port, 
-			int *query_port, unsigned long *local_ttl, char **addn_hosts, struct dhcp_context **dhcp,
-			struct dhcp_config **dhcp_conf, struct dhcp_opt **dhcp_opts, struct dhcp_vendor **dhcp_vendors, char **dhcp_file,
-			char **dhcp_sname, struct in_addr *dhcp_next_server, int *dhcp_max, 
-			unsigned int *min_leasetime, struct doctor **doctors)
+struct daemon *read_opts (int argc, char **argv)
 {
+  struct daemon *daemon = safe_malloc(sizeof(struct daemon));
+  char *problem = NULL, *buff = safe_malloc(MAXDNAME);
   int option = 0, i;
-  unsigned int flags = 0;
   FILE *file_save = NULL, *f = NULL;
   char *file_name_save = NULL, *conffile = CONFFILE;
   int conffile_set = 0;
   int line_save = 0, lineno = 0;
   opterr = 0;
   
-  *min_leasetime = UINT_MAX;
+  memset(daemon, 0, sizeof(struct daemon));
+  daemon->namebuff = buff;
 
+  /* Set defaults - everything else is zero or NULL */
+  daemon->min_leasetime = UINT_MAX;
+  daemon->cachesize = CACHESIZ;
+  daemon->port = NAMESERVER_PORT;
+  daemon->default_resolv.is_default = 1;
+  daemon->default_resolv.name = RESOLVFILE;
+  daemon->resolv_files = &daemon->default_resolv;
+  daemon->username = CHUSER;
+  daemon->groupname = CHGRP;
+  daemon->runfile =  RUNFILE;
+  daemon->dhcp_max = MAXLEASES;
+  daemon->edns_pktsz = EDNS_PKTSZ;
+  
   while (1)
     {
+      problem = NULL;
+
       if (!f)
 #ifdef HAVE_GETOPT_LONG
 	option = getopt_long(argc, argv, OPTSTRING, (struct option *)opts, NULL);
@@ -256,20 +270,23 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
      
       if (!f && option == 'w')
 	{
-	  fprintf (stderr, usage,  CACHESIZ, MAXLEASES);
+	  fprintf (stderr, usage,  CACHESIZ, EDNS_PKTSZ, MAXLEASES);
 	  exit(0);
 	}
 
       if (!f && option == 'v')
         {
-          fprintf(stderr, "dnsmasq version %s\n", VERSION);
+          fprintf(stderr, "Dnsmasq version %s  %s\n\n", VERSION, COPYRIGHT);
+	  fprintf(stderr, "This software comes with ABSOLUTELY NO WARRANTY.\n");
+	  fprintf(stderr, "Dnsmasq is free software, and you are welcome to redistribute it\n");
+	  fprintf(stderr, "under the terms of the GNU General Public License, version 2.\n");
           exit(0);
         }
       
       for (i=0; optmap[i].c; i++)
 	if (option == optmap[i].c)
 	  {
-	    flags |= optmap[i].flag;
+	    daemon->options |= optmap[i].flag;
 	    option = 0;
 	    if (f && optarg)
 	      {
@@ -287,7 +304,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 	      complain(buff, NULL);
 	      continue;
 	    }
-	               
+	      	  
 	  switch (option)
 	    { 
 	     case 'C': 
@@ -314,13 +331,13 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 	       goto fileopen;
 	      
 	    case 'x': 
-	      *runfile = safe_string_alloc(optarg);
+	      daemon->runfile = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'r':
 	      {
 		char *name = safe_string_alloc(optarg);
-		struct resolvc *new, *list = *resolv_files;
+		struct resolvc *new, *list = daemon->resolv_files;
 		if (list && list->is_default)
 		  {
 		    /* replace default resolv file - possibly with nothing */
@@ -341,71 +358,94 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		    new->logged = 0;
 		    list = new;
 		  }
-		*resolv_files = list;
+		daemon->resolv_files = list;
 		break;
 	      }
 
 	    case 'm':
-	      if (!canonicalise(optarg))
-		option = '?';
-	      else 
-		*mxname = safe_string_alloc(optarg);
-	      break;
+	      {
+		char *comma = strchr(optarg, ',');
+		if (comma)
+		  *(comma++) = 0;
+		if (!canonicalise(optarg) || (comma && !canonicalise(comma)))
+		  {
+		    option = '?';
+		    problem = "bad MX name";
+		  }
+		else 
+		  {
+		    struct mx_record *new = safe_malloc(sizeof(struct mx_record));
+		    new->next = daemon->mxnames;
+		    daemon->mxnames = new;
+		    new->mxname = safe_string_alloc(optarg);
+		    new->mxtarget = safe_string_alloc(comma); /* may be NULL */
+		  }
+		break;
+	      }
 	      
 	    case 't':
 	      if (!canonicalise(optarg))
-		option = '?';
+		{
+		  option = '?';
+		  problem = "bad MX target";
+		}
 	      else
-		*mxtarget = safe_string_alloc(optarg);
+		daemon->mxtarget = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'l':
-	      *lease_file = safe_string_alloc(optarg);
+	      daemon->lease_file = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'H':
-	      if (*addn_hosts)
-		option = '?';
+	      if (daemon->addn_hosts)
+		{
+		  option = '?';
+		  problem = "only one addn hosts file allowed";
+		}
 	      else
-		*addn_hosts = safe_string_alloc(optarg);
+		daemon->addn_hosts = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 's':
-	      if (!canonicalise(optarg))
+	      if (strcmp (optarg, "#") == 0)
+		daemon->options |= OPT_RESOLV_DOMAIN;
+	      else if (!canonicalise(optarg))
 		option = '?';
 	      else
-		*domain_suffix = safe_string_alloc(optarg);
+		daemon->domain_suffix = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'u':
-	      *username = safe_string_alloc(optarg);
+	      daemon->username = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'g':
-	      *groupname = safe_string_alloc(optarg);
+	      daemon->groupname = safe_string_alloc(optarg);
 	      break;
 	      
 	    case 'i':
 	      {
 		struct iname *new = safe_malloc(sizeof(struct iname));
-		new->next = *if_names;
-		*if_names = new;
+		new->next = daemon->if_names;
+		daemon->if_names = new;
 		/* new->name may be NULL if someone does
 		   "interface=" to disable all interfaces except loop. */
 		new->name = safe_string_alloc(optarg);
+		new->isloop = new->used = 0;
 		if (strchr(optarg, ':'))
-		  flags |= OPT_NOWILD;
+		  daemon->options |= OPT_NOWILD;
 		break;
 	      }
 	      
 	    case 'I':
 	      {
 		struct iname *new = safe_malloc(sizeof(struct iname));
-		new->next = *if_except;
-		*if_except = new;
+		new->next = daemon->if_except;
+		daemon->if_except = new;
 		new->name = safe_string_alloc(optarg);
 		if (strchr(optarg, ':'))
-		   flags |= OPT_NOWILD;
+		   daemon->options |= OPT_NOWILD;
 		break;
 	      }
 	      
@@ -415,8 +455,8 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		if ((addr.s_addr = inet_addr(optarg)) != (in_addr_t)-1)
 		  {
 		    struct bogus_addr *baddr = safe_malloc(sizeof(struct bogus_addr));
-		    baddr->next = *bogus_addr;
-		    *bogus_addr = baddr;
+		    baddr->next = daemon->bogus_addr;
+		    daemon->bogus_addr = baddr;
 		    baddr->addr = addr;
 		  }
 		else
@@ -427,7 +467,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 	    case 'a':
 	      {
 		struct iname *new = safe_malloc(sizeof(struct iname));
-		new->next = *if_addrs;
+		new->next = daemon->if_addrs;
 #ifdef HAVE_IPV6
 		if (inet_pton(AF_INET, optarg, &new->addr.in.sin_addr))
 		  {
@@ -461,7 +501,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		  }
 		
 		if (new)
-		  *if_addrs = new;
+		  daemon->if_addrs = new;
 		break;
 	      }
 	      
@@ -481,7 +521,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 			/* # matches everything and becomes a zero length domain string */
 			if (strcmp(optarg, "#") == 0)
 			  domain = "";
-			else if (!canonicalise(optarg))
+			else if (!canonicalise(optarg) && strlen(optarg) != 0)
 			  option = '?';
 			else
 			  domain = safe_string_alloc(optarg); /* NULL if strlen is zero */
@@ -534,7 +574,10 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 			  { 
 			    *portno = 0;
 			    if (!atoi_check(portno+1, &source_port))
-			      option = '?';
+			      {
+				option = '?';
+				problem = "bad port";
+			      }
 			  }
 		      }
 		    
@@ -542,7 +585,10 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		      {
 			*portno = 0;
 			if (!atoi_check(portno+1, &serv_port))
-			  option = '?';
+			  {
+			    option = '?';
+			    problem = "bad port";
+			  }
 		      }
 
 #ifdef HAVE_IPV6
@@ -614,8 +660,8 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 			serv->next->source_addr = serv->source_addr;
 			serv = serv->next;
 		      }
-		    serv->next = *serv_addrs;
-		    *serv_addrs = newlist;
+		    serv->next = daemon->servers;
+		    daemon->servers = newlist;
 		  }
 		break;
 	      }
@@ -634,18 +680,27 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		    else if (size > 10000)
 		      size = 10000;
 		    
-		    *cachesize = size;
+		    daemon->cachesize = size;
 		  }
 		break;
 	      }
 	      
 	    case 'p':
-	      if (!atoi_check(optarg, port))
+	      if (!atoi_check(optarg, &daemon->port))
 		option = '?';
 	      break;
 	      
+	    case 'P':
+	      {
+		int i;
+		if (!atoi_check(optarg, &i))
+		  option = '?';
+		daemon->edns_pktsz = (unsigned short)i;	
+		break;
+	      }
+
 	    case 'Q':
-	      if (!atoi_check(optarg, query_port))
+	      if (!atoi_check(optarg, &daemon->query_port))
 		option = '?';
 	      break;
 
@@ -655,12 +710,12 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		if (!atoi_check(optarg, &ttl))
 		  option = '?';
 		else
-		  *local_ttl = (unsigned long)ttl;
+		  daemon->local_ttl = (unsigned long)ttl;
 		break;
 	      }
 
 	    case 'X':
-	      if (!atoi_check(optarg, dhcp_max))
+	      if (!atoi_check(optarg, &daemon->dhcp_max))
 		option = '?';
 	      break;
 
@@ -670,13 +725,17 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		char *cp, *comma, *a[5] = { NULL, NULL, NULL, NULL, NULL };
 		struct dhcp_context *new = safe_malloc(sizeof(struct dhcp_context));
 		
-		new->next = *dhcp;
-		new->lease_time = DEFLEASE; 
+		new->next = daemon->dhcp;
+		new->lease_time = DEFLEASE;
+		new->addr_epoch = 0;
 		new->netmask.s_addr = 0;
 		new->broadcast.s_addr = 0;
+		new->router.s_addr = 0;
 		new->netid.net = NULL;
+		new->static_only = 0;
 		
-		
+		problem = "bad dhcp-range";
+
 		for (cp = optarg; *cp; cp++)
 		  if (!(*cp == ' ' || *cp == '.' ||  (*cp >='0' && *cp <= '9')))
 		    break;
@@ -701,22 +760,39 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		if ((k < 2) || ((new->start.s_addr = inet_addr(a[0])) == (in_addr_t)-1))
 		  option = '?';
 		else if (strcmp(a[1], "static") == 0)
-		  new->end = new->start;
+		  {
+		    new->end = new->start;
+		    new->static_only = 1;
+		  }
 		else if ((new->end.s_addr = inet_addr(a[1])) == (in_addr_t)-1)
 		  option = '?';
 		  
+		if (ntohl(new->start.s_addr) > ntohl(new->end.s_addr))
+		  {
+		    struct in_addr tmp = new->start;
+		    new->start = new->end;
+		    new->end = tmp;
+		  }
+		    
+		if (option != '?' && k >= 3 && strchr(a[2], '.') &&  
+		    ((new->netmask.s_addr = inet_addr(a[2])) != (in_addr_t)-1))
+		  {
+		    leasepos = 3;
+		    if (!is_same_net(new->start, new->end, new->netmask))
+		      {
+			problem = "inconsistent DHCP range";
+			option = '?';
+		      }
+		  }
+
 		if (option == '?')
 		  {
 		    free(new);
 		    break;
 		  }
 		else
-		  *dhcp = new;
-		
-		if (k >= 3 && strchr(a[2], '.') &&  
-		    ((new->netmask.s_addr = inet_addr(a[2])) != (in_addr_t)-1))
-		  leasepos = 3;
-		
+		  daemon->dhcp = new;
+
 		if (k >= 4 && strchr(a[3], '.') &&  
 		    ((new->broadcast.s_addr = inet_addr(a[3])) != (in_addr_t)-1))
 		  leasepos = 4;
@@ -750,8 +826,8 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		      }
 		  }
 				
-		if (new->lease_time < *min_leasetime)
-		  *min_leasetime = new->lease_time;
+		if (new->lease_time < daemon->min_leasetime)
+		  daemon->min_leasetime = new->lease_time;
 		break;
 	      }
 
@@ -763,7 +839,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		struct dhcp_config *new = safe_malloc(sizeof(struct dhcp_config));
 		struct in_addr in;
 
-		new->next = *dhcp_conf;
+		new->next = daemon->dhcp_conf;
 		new->flags = 0;		  
 		
 		
@@ -905,6 +981,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 
 		if (option == '?')
 		  {
+		    problem = "bad dhcp-host";
 		    if (new->flags & CONFIG_NAME)
 		      free(new->hostname);
 		    if (new->flags & CONFIG_CLID)
@@ -915,9 +992,9 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		  }
 		else
 		  {
-		    if ((new->flags & CONFIG_TIME) && new->lease_time < *min_leasetime)
-		      *min_leasetime = new->lease_time;
-		    *dhcp_conf = new;
+		    if ((new->flags & CONFIG_TIME) && new->lease_time < daemon->min_leasetime)
+		      daemon->min_leasetime = new->lease_time;
+		    daemon->dhcp_conf = new;
 		  }
 		break;
 	      }
@@ -928,7 +1005,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		char *cp, *comma;
 		int addrs, digs, is_addr, is_hex, is_dec;
 		
-		new->next = *dhcp_opts;
+		new->next = daemon->dhcp_opts;
 		new->len = 0;
 		new->is_addr = 0;
 		new->netid = NULL;
@@ -953,13 +1030,14 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		if ((new->opt = atoi(optarg)) == 0)
 		  {
 		    option = '?';
+		    problem = "bad dhcp-opt";
 		    if (new->netid)
 		      free(new->netid);
 		    free(new);
 		    break;
 		  }
 		
-		*dhcp_opts = new;
+		daemon->dhcp_opts = new;
 		
 		if (!comma)
 		  break;
@@ -1016,7 +1094,7 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		else if (is_dec)
 		  {
 		    /* Given that we don't know the length,
-		       this applaing hack is the best available */
+		       this appaling hack is the best available */
 		    unsigned int val = atoi(comma+1);
 		    if (val < 256)
 		      {
@@ -1074,14 +1152,14 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		
 		if ((comma = strchr(optarg, ',')))
 		  *comma = 0;
-		*dhcp_file = safe_string_alloc(optarg);
+		daemon->dhcp_file = safe_string_alloc(optarg);
 		if (comma)
 		  {
 		    optarg = comma+1;
 		    if ((comma = strchr(optarg, ',')))
 		      *comma = 0;
-		    *dhcp_sname = safe_string_alloc(optarg);
-		    if (comma && (dhcp_next_server->s_addr = inet_addr(comma+1)) == (in_addr_t)-1)
+		    daemon->dhcp_sname = safe_string_alloc(optarg);
+		    if (comma && (daemon->dhcp_next_server.s_addr = inet_addr(comma+1)) == (in_addr_t)-1)
 		      option = '?';
 		  }
 		break;
@@ -1103,8 +1181,8 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		    new->data = safe_malloc(new->len);
 		    memcpy(new->data, comma+1, new->len);
 		    new->is_vendor = (option == 'U');
-		    new->next = *dhcp_vendors;
-		    *dhcp_vendors = new;
+		    new->next = daemon->dhcp_vendors;
+		    daemon->dhcp_vendors = new;
 		  }
 		break;
 	      }
@@ -1141,8 +1219,8 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 		new->in = in;
 		new->out = out;
 		new->mask = mask;
-		new->next = *doctors;
-		*doctors = new;
+		new->next = daemon->doctors;
+		daemon->doctors = new;
 		
 		break;
 	      }
@@ -1153,62 +1231,97 @@ unsigned int read_opts (int argc, char **argv, char *buff, struct resolvc **reso
 	{
 	  if (f)
 	    {
-	      sprintf(buff, "error at line %d of %s ", lineno, conffile);
+	      sprintf(buff, "%s at line %d of %s ", 
+		      problem ? problem : "error", lineno, conffile);
 	      complain(buff, NULL);
 	    }
 	  else
-	    die("bad command line options: try --help.", NULL);
+	    die("bad command line options: %s.", problem ? problem : "try --help");
 	}
     }
       
   /* port might no be known when the address is parsed - fill in here */
-  if (*serv_addrs)
+  if (daemon->servers)
     {
       struct server *tmp;
-      for (tmp = *serv_addrs; tmp; tmp = tmp->next)
+      for (tmp = daemon->servers; tmp; tmp = tmp->next)
 	if (!(tmp->flags & SERV_HAS_SOURCE))
 	  {
 	    if (tmp->source_addr.sa.sa_family == AF_INET)
-	      tmp->source_addr.in.sin_port = htons(*query_port);
+	      tmp->source_addr.in.sin_port = htons(daemon->query_port);
 #ifdef HAVE_IPV6
 	    else if (tmp->source_addr.sa.sa_family == AF_INET6)
-	      tmp->source_addr.in6.sin6_port = htons(*query_port);
+	      tmp->source_addr.in6.sin6_port = htons(daemon->query_port);
 #endif  
 	  }
     }
   
-  if (*if_addrs)
+  if (daemon->if_addrs)
     {  
       struct iname *tmp;
-      for(tmp = *if_addrs; tmp; tmp = tmp->next)
+      for(tmp = daemon->if_addrs; tmp; tmp = tmp->next)
 	if (tmp->addr.sa.sa_family == AF_INET)
-	  tmp->addr.in.sin_port = htons(*port);
+	  tmp->addr.in.sin_port = htons(daemon->port);
 #ifdef HAVE_IPV6
 	else if (tmp->addr.sa.sa_family == AF_INET6)
-	  tmp->addr.in6.sin6_port = htons(*port);
+	  tmp->addr.in6.sin6_port = htons(daemon->port);
 #endif /* IPv6 */
     }
 		      
   /* only one of these need be specified: the other defaults to the
      host-name */
-  if ((flags & OPT_LOCALMX) || *mxname || *mxtarget)
+  if ((daemon->options & OPT_LOCALMX) || daemon->mxnames || daemon->mxtarget)
     {
       if (gethostname(buff, MAXDNAME) == -1)
 	die("cannot get host-name: %s", NULL);
 	      
-      if (!*mxname)
-	*mxname = safe_string_alloc(buff);
+      if (!daemon->mxnames)
+	{
+	  daemon->mxnames = safe_malloc(sizeof(struct mx_record));
+	  daemon->mxnames->next = NULL;
+	  daemon->mxnames->mxtarget = NULL;
+	  daemon->mxnames->mxname = safe_string_alloc(buff);
+}
       
-      if (!*mxtarget)
-	*mxtarget = safe_string_alloc(buff);
+      if (!daemon->mxtarget)
+	daemon->mxtarget = safe_string_alloc(buff);
     }
   
-  if (flags & OPT_NO_RESOLV)
-    *resolv_files = 0;
-  else if (*resolv_files && (*resolv_files)->next && (flags & OPT_NO_POLL))
+  if (daemon->options & OPT_NO_RESOLV)
+    daemon->resolv_files = 0;
+  else if (daemon->resolv_files && (daemon->resolv_files)->next && (daemon->options & OPT_NO_POLL))
     die("only one resolv.conf file allowed in no-poll mode.", NULL);
+  
+  if (daemon->options & OPT_RESOLV_DOMAIN)
+    {
+      char *line;
+      
+      if (!daemon->resolv_files || (daemon->resolv_files)->next)
+	die("must have exactly one resolv.conf to read domain from.", NULL);
+      
+      if (!(f = fopen((daemon->resolv_files)->name, "r")))
+	die("failed to read %s: %m", (daemon->resolv_files)->name);
+      
+      while ((line = fgets(buff, MAXDNAME, f)))
+	{
+	  char *token = strtok(line, " \t\n\r");
+	  
+	  if (!token || strcmp(token, "search") != 0)
+	    continue;
+	  
+	  if ((token = strtok(NULL, " \t\n\r")) &&  
+	      canonicalise(token) &&
+	      (daemon->domain_suffix = safe_string_alloc(token)))
+	    break;
+	}
 
-  return flags;
+      fclose(f);
+
+      if (!daemon->domain_suffix)
+	die("no search directive found in %s", (daemon->resolv_files)->name);
+    }
+      
+  return daemon;
 }
       
       
